"""
Convert a umls release to a jsonl file of concepts. Originally coming from https://github.com/allenai/scispacy/blob/main/scripts/export_umls_json.py
"""
import csv
from dataclasses import dataclass
import argparse
import os
import pathlib
from typing import List
import pandas as pd
from neo4j import GraphDatabase, WRITE_ACCESS
from functools import reduce


def getNeo4jImportPath():
    return (
        pathlib.Path(
            "/Users/coltonflowers/Library/Application Support/Neo4j Desktop/Application/relate-data/dbmss/dbms-59ffefaf-0a98-4769-9e80-856f2940c97a"
        )
        / "import"
    )


NEO4J_UI = "bolt://localhost:7687"
DRIVER = GraphDatabase.driver(NEO4J_UI, auth=("neo4j", "password"))
CSV_DIRECTORY = getNeo4jImportPath()


def read_umls_file_headers(meta_path: pathlib.Path, filename: str) -> List[str]:
    """
    Read the file descriptor MRFILES.RRF from a UMLS release and get column headers (names)
    for the given file
    MRFILES.RRF file format: a pipe-separated values
    Useful columns:
        column 0: name of one of the files in the META directory
        column 2: column names of that file
    Args:
        meta_path: path to the META directory of an UMLS release
        filename: name of the file to get its column headers
    Returns:
        a list of column names
    """
    file_descriptors = f"{meta_path}/MRFILES.RRF"  # to get column names
    with open(file_descriptors, encoding="utf-8") as fin:
        for line in fin:
            splits = line.split("|")
            found_filename = splits[0]
            column_names = (splits[2] + ",").split(
                ","
            )  # ugly hack because all files end with an empty column
            if found_filename in filename:
                # column_names.remove('')
                return column_names
    assert False, f"Couldn't find column names for file {filename}"
    return None


def read_umls_col_data_types(meta_path: pathlib.Path, filename: str) -> List[str]:
    """
    Read the file descriptor MRFILES.RRF from a UMLS release and get column headers (names)
    for the given file
    MRFILES.RRF file format: a pipe-separated values
    Useful columns:
        column 0: name of one of the files in the META directory
        column 2: column names of that file
    Args:
        meta_path: path to the META directory of an UMLS release
        filename: name of the file to get its column headers
    Returns:
        a list of column names
    """
    cols_filename = f"{meta_path}/MRCOLS.RRF"  # to get column names
    headers = read_umls_file_headers(meta_path, cols_filename)
    data_types = {}
    with open(cols_filename, encoding="utf-8") as fin:
        for line in fin:
            splits = line.strip().split("|")
            assert len(headers) == len(splits)
            record = dict(zip(headers, splits))
            if record["FIL"] == filename:
                data_types[record["COL"]] = record["DTY"]
    return [
        data_types[header]
        for header in read_umls_file_headers(meta_path, filename)
        if header
    ]


def process_file(meta_path, headers, filename, attrs, csv_file_name):
    with open(f"{meta_path}/{filename}", encoding="utf-8") as fin:
        pd.read_csv(
            fin, delimiter="|", usecols=attrs, names=headers
        ).drop_duplicates().replace(to_replace=r"\\", value="", regex=True).to_csv(
            os.path.join(CSV_DIRECTORY, csv_file_name)
        )


def get_cypher_type_converter_string(attr, data_types):
    if data_types[attr] == "integer":
        type = f"toInteger(line.{attr})"
    else:
        type = f"line.{attr}"
    if attr == "RELA":
        return f"coalesce({type}, 'NONE')"
    else:
        return type


def import_nodes(meta_path, filename, node_name, attrs, index_attr):
    print(f"Creating {node_name} nodes...")
    csv_file_name = f"{node_name}.csv"
    headers = read_umls_file_headers(meta_path, filename)
    data_types = dict(zip(headers, read_umls_col_data_types(meta_path, filename)))
    # process_file(meta_path,headers,filename,attrs,csv_file_name)
    query_string = create_node_query_string(node_name, attrs, csv_file_name, data_types)
    print(query_string)
    with DRIVER.session(default_access_mode=WRITE_ACCESS) as session:
        session.run(query_string)
        if node_name == "definition":
            session.run(
                f"CREATE INDEX {node_name}_index FOR (t:{node_name}) ON (t.{index_attr})"
            ).consume()
        else:
            session.run(
                f"CREATE CONSTRAINT ON (t:{node_name}) ASSERT t.{index_attr} IS UNIQUE;"
            )


def create_node_query_string(node_name, attrs, csv_file_name, data_types):

    attr_string = str(
        {attr: get_cypher_type_converter_string(attr, data_types) for attr in attrs}
    )
    stringified_attrs = ", ".join(
        [
            f"{attr}: {get_cypher_type_converter_string(attr, data_types)}"
            for attr in attrs
        ]
    )
    attr_string = f"{{{stringified_attrs}}}"
    query = """
            USING PERIODIC COMMIT
            LOAD CSV WITH HEADERS FROM '{0}' AS line
            CREATE (:{1} {2})
            """.format(
        f"file:///{csv_file_name}", node_name, attr_string
    )
    return query


@dataclass
class NodeBridge:
    name: str
    node_attr: str
    rel_attr: str


def create_edge_query_string(
    node_bridge_one,
    node_bridge_two,
    edge_properties,
    csv_file_name,
    data_types,
    transaction_number,
):
    attr_string_1 = f"{{{node_bridge_one.node_attr}:{get_cypher_type_converter_string(node_bridge_one.rel_attr, data_types)}}}"
    attr_string_2 = f"{{{node_bridge_two.node_attr}:{get_cypher_type_converter_string(node_bridge_two.rel_attr, data_types)}}}"
    stringified_edge_properties = [
        f"{attr}:{get_cypher_type_converter_string(attr, data_types)}"
        for attr in edge_properties
        if not attr in [node_bridge_one.rel_attr, node_bridge_two.rel_attr]
    ]
    if stringified_edge_properties:
        edge_property_str = "{"
        edge_property_str += ", ".join(stringified_edge_properties)
        edge_property_str += "}"
    else:
        edge_property_str = ""

    query = """
            LOAD CSV WITH HEADERS FROM '{0}' AS line
            CALL {{ 
            WITH line
            MATCH (n1:{1} {2})
            MATCH (n2:{3} {4})
            MERGE (n1)-[:{5} {6}]->(n2)
            }} IN TRANSACTIONS OF {7} ROWS
            """.format(
        f"file:///{csv_file_name}",
        node_bridge_one.name,
        attr_string_1,
        node_bridge_two.name,
        attr_string_2,
        f"{node_bridge_one.name}_to_{node_bridge_two.name}",
        edge_property_str,
        transaction_number,
    )
    return query


def import_edges(
    meta_path,
    filename,
    node_bridge_one: NodeBridge,
    node_bridge_two: NodeBridge,
    edge_properties,
    transaction_number=10000,
):
    print(f"Creating {node_bridge_one.name}-{node_bridge_two.name} edges...")
    csv_file_name = f"{node_bridge_one.name}_to_{node_bridge_two.name}.csv"
    headers = read_umls_file_headers(meta_path, filename)
    data_types = dict(zip(headers, read_umls_col_data_types(meta_path, filename)))
    # process_file(meta_path,headers,filename,edge_properties,csv_file_name)
    query_string = create_edge_query_string(
        node_bridge_one,
        node_bridge_two,
        edge_properties,
        csv_file_name,
        data_types,
        transaction_number,
    )
    print(query_string)
    with DRIVER.session(default_access_mode=WRITE_ACCESS) as session:
        session.run(query_string)


def port_umls_to_neo4j(meta_path):
    import_nodes(
        meta_path, "MRDEF.RRF", "definition", ["SAB", "DEF", "SUPPRESS", "AUI"], "AUI"
    )
    import_nodes(meta_path, "MRCONSO.RRF", "concept", ["CUI"], "CUI")
    import_nodes(
        meta_path,
        "MRCONSO.RRF",
        "atom",
        ["AUI", "SUPPRESS", "CVF", "SRL", "SAB"],
        "AUI",
    )
    import_nodes(meta_path, "MRCONSO.RRF", "string", ["SUI", "STR"], "SUI")
    import_nodes(meta_path, "MRCONSO.RRF", "term", ["LUI", "LAT"], "LUI")
    import_nodes(meta_path, "MRSTY.RRF", "semantic_type", ["STY", "TUI", "STN"], "TUI")

    import_edges(
        meta_path,
        "MRREL.RRF",
        NodeBridge("atom", "AUI", "AUI2"),
        NodeBridge("atom", "AUI", "AUI1"),
        ["SAB", "AUI1", "AUI2", "REL", "RELA", "RUI", "SUPPRESS"],
        transaction_number=1000,
    )
    import_edges(
        meta_path,
        "MRCONSO.RRF",
        NodeBridge("concept", "CUI", "CUI"),
        NodeBridge("atom", "AUI", "AUI"),
        ["CUI", "AUI", "ISPREF", "TS", "STT", "TTY"],
        transaction_number=1000,
    )
    import_edges(
        meta_path,
        "MRCONSO.RRF",
        NodeBridge("string", "SUI", "SUI"),
        NodeBridge("atom", "AUI", "AUI"),
        ["AUI", "SUI"],
        transaction_number=1000,
    )
    import_edges(
        meta_path,
        "MRCONSO.RRF",
        NodeBridge("string", "SUI", "SUI"),
        NodeBridge("term", "LUI", "LUI"),
        ["SUI", "LUI"],
        transaction_number=500,
    )

    import_edges(
        meta_path,
        "MRSTY.RRF",
        NodeBridge("concept", "CUI", "CUI"),
        NodeBridge("semantic_type", "TUI", "TUI"),
        ["CUI", "TUI"],
        transaction_number=1000,
    )
    import_edges(
        meta_path,
        "MRDEF.RRF",
        NodeBridge("atom", "AUI", "AUI"),
        NodeBridge("definition", "AUI", "AUI"),
        ["SAB", "SUPPRESS", "AUI"],
        transaction_number=10000,
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--meta_path", help="Path to the META directory of an UMLS release."
    )
    args = parser.parse_args()
    # process_file(args.meta_path,filename="MRCONSO.RRF",attrs=["SUI", "STR"],csv_file_name="string.csv")
    port_umls_to_neo4j(args.meta_path)
