import csv
from typing import Tuple
from crtl.lkb.modify_snomed_hierarchy import (
    get_snomed_preferred_hierarchy,
    get_snomed_preferred_nodes,
)
from neo4j import GraphDatabase
import networkx as nx

NEO4J_UI = "bolt://localhost:7687"
DRIVER = GraphDatabase.driver(NEO4J_UI, auth=("neo4j", "password"))


def get_snomed_filtered_auis() -> set[str]:
    snomed_preferred_nodes = get_snomed_preferred_nodes()
    return set(node["aui"] for node in snomed_preferred_nodes if not node["removed"])


def get_rxnorm_bn_links() -> set[Tuple[str, str, str]]:
    snomed_filtered_atoms = get_snomed_filtered_auis()
    with DRIVER.session() as session:
        query_string = """
        UNWIND $snomed_filtered_atoms as aui
        MATCH (brandname_a:atom{SAB: "RXNORM", SUPPRESS: "N"})-[:atom_to_atom{RELA:"tradename_of"}]->(generic_a:atom{SAB: "RXNORM", SUPPRESS: "N"})<-[:concept_to_atom]-(:concept)--(a1:atom{AUI:aui}) 
        RETURN brandname_a.AUI as a1, generic_a.AUI as a2
        """
        records = session.run(
            query_string,
            parameters={"snomed_filtered_atoms": list(snomed_filtered_atoms)},
        )
        return set(
            (record["a1"], "tradename_of", record["a2"]) for record in records
        ).union(
            set((record["a2"], "has_tradename", record["a1"]) for record in records)
        )


def get_linkable_brand_names_and_generic_rxnorm_auis() -> set[str]:
    records = get_rxnorm_bn_links()
    brandname_auis = set([record[0] for record in records])
    generic_auis = set([record[2] for record in records])
    return brandname_auis.union(generic_auis)


def get_nonhierarcical_rels():
    snomed_filtered_atoms = get_snomed_filtered_auis()
    with DRIVER.session() as session:
        ##NOTE FOR IMPROVEMENT: READ FROM THE FILTERED SNOMED_PREFERRED_ATOMS?
        # GET ALL NON-HIERARCHICAL RELATIONS AMONG PREFERRED TERM SNOMED ATOM
        query_string = """
    UNWIND $snomed_filtered_atoms as aui
    MATCH(a1:atom{AUI:aui,SAB:"SNOMEDCT_US",SUPPRESS:"N"})-[r:atom_to_atom{SAB:"SNOMEDCT_US",SUPPRESS:"N"}]->(a2:atom{SAB:"SNOMEDCT_US",SUPPRESS:"N"})
    WHERE a1 in trendnet_atoms AND a2 in trendnet_atoms AND r.RELA not in ["mapped_to","isa","inverse_isa","mapped_from","inverse_was_a","same_as","replaced_by","was_a","possibly_equivalent_to"]
    RETURN a1,r.RELA as rela, a2
    """
        return set(
            (record["a1"], record["rela"], record["a2"])
            for record in session.run(
                query_string,
                parameters={"snomed_filtered_atoms": list(snomed_filtered_atoms)},
            )
        )


def get_trendnet_auis() -> set[str]:
    filtered_snomed_preferred_auis = get_snomed_filtered_auis()
    rxnorm_auis = get_linkable_brand_names_and_generic_rxnorm_auis()
    return filtered_snomed_preferred_auis.union(rxnorm_auis)


def get_hierarchical_rels() -> set[Tuple[str, str, str]]:
    snomed_preferred_hierarchy = get_snomed_preferred_hierarchy()
    return set((aui1, "isa", aui2) for aui1, aui2 in snomed_preferred_hierarchy).union(
        (aui2, "inverse_isa", aui1) for aui1, aui2 in snomed_preferred_hierarchy
    )


def get_trendnet_atom_rels() -> set[Tuple[str, str, str]]:
    hierarchical_rels = get_hierarchical_rels()
    nonhiercichal_rels = get_nonhierarcical_rels()
    tradename_rels = get_rxnorm_bn_links()
    return hierarchical_rels.union(nonhiercichal_rels).union(tradename_rels)


def get_trendnet_cuis() -> set[str]:
    trendnet_auis = get_trendnet_auis()
    print(len(trendnet_auis))
    with DRIVER.session() as session:
        query_string = """
        UNWIND $trendnet_auis as aui
        MATCH (a:atom{AUI:aui})--(c:concept)
        RETURN distinct c.CUI as cui
        """
        return set(
            record["cui"]
            for record in session.run(
                query_string, parameters={"trendnet_auis": list(trendnet_auis)}
            )
        )


def get_trendnet_terms() -> set[Tuple[str, str, str]]:
    trendnet_cuis = get_trendnet_cuis()
    print(len(trendnet_cuis))
    with DRIVER.session() as session:
        query_string = """
        UNWIND $trendnet_cuis as cui
        MATCH (l:term{LAT:"ENG"})--(s:string)--(a:atom{SUPPRESS:"N"})-[r:concept_to_atom]-(c:concept{CUI:cui})
        WHERE ((a.SAB in ["SNOMEDCT_US"] AND r.TTY IN ["PT","SY"]) OR (a.SAB in ["RXNORM","MSH"]))
        RETURN cui,l.LUI as lui,s.STR as string
        """
        records = set(
            [
                (record["cui"], record["lui"], record["string"])
                for record in session.run(
                    query_string, parameters={"trendnet_cuis": list(trendnet_cuis)}
                )
            ]
        )
    with open("norm_input.data", "w") as f:
        writer = csv.DictWriter(f, fieldnames=["cui", "lui", "string"], delimiter="|")
        writer.writerows(
            [
                {"cui": record[0], "lui": record[1], "string": record[2]}
                for record in records
            ]
        )


if __name__ == "__main__":
    get_trendnet_terms()

##1) GET RELEVANT TERMS FROM GOOD SOURCE/TTY COMBOS
##2) GET ALL STRINGS FOR THESE TERMS
##3) RUN THEM THROUGH THE LEXICAL NORM PROGRAM


# GET RELATIONS AMONG TRENDNET CONCEPTS FROM ALL TRENDNET ATOM RELATIONS
def get_trendnet_concept_rels():
    """
    UNWIND $trendnet_atom_rels as atom_rels
    WITH atom_rels.aui1 as aui1, atom_rels.RELA as rela, atom_rels.aui2
    MATCH (a:atom{AUI:au2})--(c2:concept)
    MATCH (c1:concept)--(a:atom{AUI:aui1})
    RETURN c1.CUI,rela,c2.CUI
    """
    trendnet_atom_rels = get_trendnet_atom_rels()


# GET ALL SEMANTIC TYPES FOR TRENDNET CONCEPTS
"""
CALL{
MATCH (a1:atom{SAB: "SNOMEDCT_US", SUPPRESS: "N"})<-[:concept_to_atom{TTY:"PT"}]-(c1:concept)--(aa1:atom{SUPPRESS: "N"}) 
WITH a1,collect(distinct aa1.SAB) as related_atom_sources
WITH a1,NOT (("CHV" IN related_atom_sources OR "MSH" IN related_atom_sources) OR (a1.AUI in ["A3291280","A10887814","A7880728","A10884570","A32314570","A3591013","A3684559"]) OR ((a1:atom{SAB: "SNOMEDCT_US",SUPPRESS: "N"})-[:atom_to_atom{RELA:"isa",SAB:"SNOMEDCT_US",SUPPRESS:"N"}]->(:atom{AUI:"A3684559"}))) as removed
WHERE NOT removed
MATCH (brandname_a:atom{SAB: "RXNORM", SUPPRESS: "N"}) -[:atom_to_atom{RELA:"tradename_of"}]->(generic_a:atom{SAB: "RXNORM", SUPPRESS: "N"})<-[:concept_to_atom]-(:concept)--(a1) 
WITH collect(distinct brandname_a) + collect(distinct generic_a) as rxnorm_atoms
UNWIND rxnorm_atoms as rxnorm_atom
RETURN distinct rxnorm_atom as atom
UNION 
MATCH (a1:atom{SAB: "SNOMEDCT_US", SUPPRESS: "N"})<-[:concept_to_atom{TTY:"PT"}]-(c1:concept)--(aa1:atom{SUPPRESS: "N"}) 
WITH a1,collect(distinct aa1.SAB) as related_atom_sources
WITH a1,NOT (("CHV" IN related_atom_sources OR "MSH" IN related_atom_sources) OR (a1.AUI in ["A3291280","A10887814","A7880728","A10884570","A32314570","A3591013","A3684559"]) OR ((a1:atom{SAB: "SNOMEDCT_US",SUPPRESS: "N"})-[:atom_to_atom{RELA:"isa",SAB:"SNOMEDCT_US",SUPPRESS:"N"}]->(:atom{AUI:"A3684559"}))) as removed
WHERE removed
RETURN distinct a1 as atom}
WITH atom
MATCH (atom)--(c:concept)
WITH distinct c
MATCH (c)--(st:semantic_type)
RETURN *
"""
