import csv
from neo4j import GraphDatabase
import networkx as nx

NEO4J_UI = "bolt://localhost:7687"
DRIVER = GraphDatabase.driver(NEO4J_UI, auth=("neo4j", "password"))


def get_snomed_preferred_nodes():
    # GET ALL PREFERRED TERM SNOMED ATOM FOR GETTING HIERARCHY
    with DRIVER.session() as session:
        get_nodes_query_string = """
            MATCH (a1:atom{SAB: "SNOMEDCT_US", SUPPRESS: "N"})--(s:string)
            OPTIONAL MATCH (a1)<-[:concept_to_atom{TTY:"PT"}]-(c1:concept)--(aa1:atom{SUPPRESS: "N"}) 
            WITH a1,s,collect(distinct aa1.SAB) as related_atom_sources
            WITH a1,s,NOT (("CHV" IN related_atom_sources OR "MSH" IN related_atom_sources) OR (a1.AUI in ["A3291280","A10887814","A7880728","A10884570","A32314570","A3591013","A3684559"]) OR ((a1:atom{SAB: "SNOMEDCT_US",SUPPRESS: "N"})-[:atom_to_atom{RELA:"isa",SAB:"SNOMEDCT_US",SUPPRESS:"N"}]->(:atom{AUI:"A3684559"}))) as removed
            RETURN a1.AUI as aui,s.STR as string,removed
            """
        return [dict(record) for record in session.run(get_nodes_query_string)]


def get_snomed_preferred_hierarchy():
    # GET ALL PREFERRED TERM SNOMED ATOM FOR GETTING HIERARCHY
    with DRIVER.session() as session:
        # GET ALL HIERARCHICAL RELATIONS AMONG PREFERRED TERM SNOMED ATOM FOR GETTING HIERARCHY
        get_rels_query_string = """
            MATCH(a1:atom{SAB:"SNOMEDCT_US",SUPPRESS:"N"})-[r:atom_to_atom{RELA:"isa",SAB:"SNOMEDCT_US",SUPPRESS:"N"}]->(a2:atom{SAB:"SNOMEDCT_US",SUPPRESS:"N"})
            RETURN a1.AUI as aui1,a2.AUI as aui2
        """
        return session.run(get_rels_query_string)


def get_snomed_hierarchy() -> nx.DiGraph:
    node_records = get_snomed_preferred_nodes()
    G = nx.DiGraph()
    for node in node_records:
        G.add_node(node["aui"], string=node["string"], removed=node["removed"])
    rel_records = get_snomed_hierarchy()
    for rel in rel_records:
        G.add_edge(rel["aui1"], rel["aui2"])
    for cycle in nx.simple_cycles(G):
        G.remove_edge(cycle[-1], cycle[0])
    return G


def get_severed_ancestors(G: nx.DiGraph) -> nx.DiGraph:
    subgraph = nx.subgraph_view(
        G, filter_edge=lambda n1, n2: G.nodes[n1]["removed"] or G.nodes[n2]["removed"]
    )
    print("\t\tGETTING CLOSURE...")
    closure = nx.transitive_closure_dag(subgraph)
    print("\t\tGETTING CLOSURE:DONE")
    subgraph = nx.subgraph_view(
        closure, filter_node=lambda n: not G.nodes[n]["removed"]
    )
    print("\t\tGETTING REDUCTION...")
    # reduction = nx.transitive_reduction(subgraph)
    print("\t\tGETTING REDUCTION:DONE")
    return subgraph


def get_modified_hierarchy(G: nx.DiGraph) -> nx.DiGraph:
    print("\tGETTING SEVERED ANCESTORS...")
    severed_ancestors = get_severed_ancestors(G)
    print("\tGETTING SEVERED ANCESTORS:DONE")
    print("\tADDING SEVERED ANCESTORS TO FILTERED HIERARCHY...")
    new_G = nx.DiGraph()
    for node in G:
        if not G.nodes[node]["removed"]:
            new_G.add_node(node, string=G.nodes[node]["string"])

    for n, nbrsdict in G.adjacency():
        for nbr, _ in nbrsdict.items():
            if not G.nodes[nbr]["removed"] and not G.nodes[n]["removed"]:
                new_G.add_edge(n, nbr)

    for n, nbrsdict in severed_ancestors.adjacency():
        for nbr, _ in nbrsdict.items():
            new_G.add_edge(n, nbr)
    print("\tADDING SEVERED ANCESTORS TO FILTERED HIERARCHY...")
    print("\tGETTING 2nd REDUCTION")
    reduction = nx.transitive_reduction(new_G)
    reduction.add_nodes_from(new_G.nodes(data=True))
    return reduction


def create_modified_hierarchy_csv():
    print("GETTING SNOMED HIERARCHY...")
    snomed_hierarchy: nx.DiGraph = get_snomed_hierarchy()
    print("GETTING SNOMED HIERARCHY:DONE")
    print("GETTING MODIFIED HIERARCHY...")
    modified_hierarchy = get_modified_hierarchy(snomed_hierarchy)
    print("\tGETTING 2nd REDUCTION:DONE")
    print("GETTING MODIFIED HIERARCHY:DONE")
    with open("modified_hierarchy.csv", "w") as f:
        writer = csv.writer(f)
        writer.writerow(["AUI1", "AUI2"])
        for n, nbrsdict in modified_hierarchy.adjacency():
            for nbr, _ in nbrsdict.items():
                writer.writerow((n, nbr))


if __name__ == "__main__":
    create_modified_hierarchy_csv()
